setup = '''
import setuptools, os

readme_path = 'README.md'

if os.path.exists(readme_path):
    with open(readme_path, 'r') as f:
        long_description = f.read()
else:
    long_description = '[PACKAGE_NAME]'

setuptools.setup(
    name='[PACKAGE_NAME]',
    version='[PACKAGE_VERSION]',
    author='[AUTHOR]',
    description='[SHORT_DESCRIPTION]',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='[GIT_URL]',
    packages=setuptools.find_packages(),
    install_requires=[[DEPENDENCIES]],
    classifiers=[
        [PYTHON_CLASSIFIERS],
        'Operating System :: OS Independent',
    ],
    python_requires='>=[MIN_PYTHON_VERSION]',
)
'''.strip()



from .api import new_api
from .class_ import new_class
from .enum_ import new_enum
from .license import new_license
from .readme import new_readme, updated_readme
from .setup import new_setup, updated_setup, current_version_number
from .install_dependencies import new_install_dependencies_file

from .core_texts import file, flow, gitignore
from .core_texts import file as new_file
from .core_texts import flow as new_flow
from .core_texts import gitignore as new_gitignore



# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, List
import os

# Pip
from kdependencies import InstalledPackage
from kcu import strings, strio

# Local
from .core_texts import setup
from .utils import multi_replace

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------ Public methods ------------------------------------------------------------ #

def updated_setup(
    old_setup: str,
    dependencies: Optional[List[str]] = None,
    version_bump_type: int = 0
) -> str:
    if dependencies:
        try:
            old_dependencies_str = old_setup.split('install_requires')[-1].split('[')[1].split(']')[0]

            if old_dependencies_str:
                old_setup = old_setup.replace(old_dependencies_str, __dependencies_str(dependencies))
        except Exception as e:
            print(e)

    try:
        old_version_str = __extract_current_version_number(old_setup)

        if old_version_str:
            old_setup = old_setup.replace(old_version_str, __bumped_version_number(old_version_str))
    except Exception as e:
        print(e)

    return old_setup

def current_version_number(setup_file_path: str) -> Optional[str]:
    if not os.path.exists(setup_file_path):
        return None

    try:
        return __extract_current_version_number(strio.load(setup_file_path))
    except:
        return None

def new_setup(
    package_name: str,
    min_python_version: Optional[float] = None,
    max_python_version: Optional[float] = None,
    package_version: Optional[str] = None,
    author: Optional[str] = None,
    git_url: Optional[str] = None,
    dependencies: Optional[List[InstalledPackage]] = None,
    short_description: Optional[str] = None,
    license_str: Optional[str] = None
) -> str:
    min_python_version = min_python_version or 3.4
    max_python_version = max_python_version or 3.9

    return multi_replace(setup, {
        '[PACKAGE_NAME]': package_name,
        '[AUTHOR]': author or '',
        '[GIT_URL]': git_url or '',
        '[PACKAGE_VERSION]': package_version or '0.0.0',
        '[DEPENDENCIES]': __dependencies_str(dependencies) if dependencies else '',
        '[SHORT_DESCRIPTION]': short_description or package_name,
        '[PYTHON_CLASSIFIERS]': __get_python_classifiers(min_python_version, max_python_version, license_str),
        '[MIN_PYTHON_VERSION]': min_python_version,
    })


# ----------------------------------------------------------- Private methods ------------------------------------------------------------ #

def __extract_current_version_number(setup_file_str: str) -> Optional[str]:
    try:
        return setup_file_str.split('version=')[-1].split('version =')[-1].split(',')[0].replace('"', '').replace("'", '')
    except:
        return None

def __get_python_classifiers(
    min_python_version: float,
    max_python_version: float,
    license_str: Optional[str] = 'License :: OSI Approved :: MIT License'
) -> str:
    current_version = min_python_version
    classifiers = [current_version]

    while current_version < max_python_version:
        current_version += 0.1
        classifiers.append('Programming Language :: Python :: {}'.format(float(int(current_version*10))/10))

    if license_str:
        classifiers.append(license_str)

    return ',\n{}'.format(8*' ').join(['\'{}\''.format(c) for c in classifiers])

def __dependencies_str(dependencies: List[InstalledPackage]) -> str:
    return '\n{}{}\n{}'.format(
        8*' ',
        ',\n{}'.format(8*' ').join(['\'{}\''.format(d.get_install_name()) for d in dependencies]),
        4*' '
    ) if dependencies else ''

def __bumped_version_number(version_number: str, bump_type: int = 0) -> str:
    comps = version_number.split('.')
    bump_index = len(comps)-1-bump_type
    comps[bump_index] = str(int(comps[bump_index]) + 1)

    return '.'.join(comps)


# ---------------------------------------------------------------------------------------------------------------------------------------- #



# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import List, Optional, Tuple, Union

# Pip
from bullet import SlidePrompt, Bullet, Input, colors

# Local
from .constants import Constants

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------ class: Prompt ------------------------------------------------------------- #

class Prompt:

    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    @classmethod
    def min_default_python_version(cls) -> float:
        return cls.__prompt((cls.__get_min_default_python_version_prompt(), float))

    @classmethod
    def max_default_python_version(cls) -> float:
        return cls.__prompt((cls.__get_max_default_python_version_prompt(), float))

    @classmethod
    def author(cls, default_username: Optional[str] = None) -> Optional[str]:
        return cls.__prompt((cls.__get_author_prompt(), str)) or default_username

    @classmethod
    def config(
        cls,
        default_author: Optional[str] = None,
        default_git_message: Optional[str] = None
    ) -> Tuple[Optional[str], float, float]:
        author, git_message, min_v, max_v = cls.__prompt([
            (cls.__get_author_prompt(default_author=default_author), str),
            (cls.__get_default_git_message_prompt(default_message=default_git_message), str),
            (cls.__get_min_default_python_version_prompt(), float),
            (cls.__get_max_default_python_version_prompt(), float)
        ])

        return author or default_author, git_message or default_git_message, min_v, max_v

    @classmethod
    def new_package(
        cls,
        passed_package_name: Optional[str] = None,
        default_package_name: Optional[str] = None
    ) -> Tuple[str, str]:
        get_package_description_prompt = (cls.__get_package_description_prompt(), str),
        # get_max_default_python_version_prompt = (cls.__get_max_default_python_version_prompt(), float)

        if passed_package_name:
            package_name, description = cls.__prompt([
                (cls.__get_package_name_prompt(default_package_name), str),
                get_package_description_prompt,
                # get_max_default_python_version_prompt
            ])
        else:
            description = cls.__prompt([
                get_package_description_prompt,
                # get_max_default_python_version_prompt
            ])[0]

        return package_name, description#, max_v


    # ------------------------------------------------------- Private methods -------------------------------------------------------- #

    # Prompts
    @staticmethod
    def __prompt(
        prompts_with_types: Union[List[Tuple[Union[Bullet, Input], any]], Tuple[Union[Bullet, Input], any]],
        summarize: bool = True
    ) -> Union[List[any], any]:
        if not isinstance(prompts_with_types, list):
            prompts_with_types = [prompts_with_types]

        prompts = [p[0] for p in prompts_with_types]
        types = [p[1] for p in prompts_with_types]

        cli = SlidePrompt(prompts)
        all_res = cli.launch()
        print(all_res)
        all_res = [res[1] if not isinstance(res[1], tuple) else res[1][0] for res in all_res]
        print(all_res)
        results = [res if isinstance(res, types[i]) else types[i](res) for i, res in enumerate(all_res)]

        if summarize:
            cli.summarize()

        return results if len(results) > 1 else results[0]


    # Getters

    @staticmethod
    def __get_package_name_prompt(default_package_name: Optional[str] = None) -> Input:
        return Input(
            "Enter package name (will be used on pip when published): ",
            default=default_package_name,
            word_color=colors.foreground["yellow"]#,
            # pattern='.*'
        )

    @staticmethod
    def __get_package_description_prompt(default_package_description: Optional[str] = None) -> Input:
        return Input(
            "Enter package description: ",
            default=default_package_description,
            word_color=colors.foreground["yellow"],
            pattern='.*'
        )

    @staticmethod
    def __get_author_prompt(default_author: Optional[str] = None) -> Input:
        return Input(
            "Default author to use for packages: ",
            default=default_author,
            word_color=colors.foreground["yellow"],
            pattern='.*'
        )

    @staticmethod
    def __get_default_git_message_prompt(default_message: Optional[str] = None) -> Input:
        return Input(
            "Default message to use for git commits: ",
            default=default_message,
            word_color=colors.foreground["yellow"],
            pattern='.*'
        )

    @classmethod
    def __get_min_default_python_version_prompt(cls) -> Bullet:
        return cls.__get_bullet_prompt('Minimum default supported python vesion?', choices=Constants.PYTHON_VERSIONS)

    @classmethod
    def __get_max_default_python_version_prompt(cls) -> Bullet:
        return cls.__get_bullet_prompt('Maximum default supported python vesion?', choices=Constants.PYTHON_VERSIONS[::-1])

    @staticmethod
    def __get_bullet_prompt(
        question: str,
        choices: List[str]
    ) -> Bullet:
        return Bullet(
            question,
            choices=choices,
            bullet=">",
            margin=2,
            bullet_color=colors.foreground["black"],
            word_color=colors.foreground["cyan"],
            word_on_switch=colors.foreground["black"],
            background_on_switch=colors.background["cyan"],
            indent=0,
            shift=1,
            pad_right=1,
            return_index=True
        )


# ---------------------------------------------------------------------------------------------------------------------------------------- #



# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, Union, Tuple, List
import os, sys

# Pip
from kcu.sh import sh
from kcu import strio, kjson, kpath

# Local
from .constants import Constants
from .prompt import Prompt
from .config import Config

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------- class: Utils ------------------------------------------------------------- #

class Utils:

    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    # CLI

    @staticmethod
    def vscode_open(path: str):
        sh('code --reuse-window {}'.format(path))

    # Python

    @staticmethod
    def python(cmd: str, debug: bool = True) -> str:
        return sh('{} {}'.format(sys.executable, cmd), debug=debug)

    @classmethod
    def pip(cls, cmd: str, debug: bool = True) -> str:
        return cls.python('-m pip {}'.format(cmd), debug=debug)

    @classmethod
    def publish(cls) -> None:
        cls.cleanup()

        # Ensure you have the latest versions of twine, setuptools, and wheel
        cls.pip('install --user --upgrade twine setuptools wheel')
        # Build the package
        cls.python('setup.py sdist bdist_wheel')
        # Upload it to pypi
        cls.python('-m twine upload --repository pypi dist/*')

        cls.cleanup()

    @staticmethod
    def cleanup(root_folder_path: Optional[str] = None) -> None:
        root_folder_path = root_folder_path or os.getcwd()

        for (_, dir_paths, _) in os.walk(root_folder_path):
            for dir_path in dir_paths:
                if dir_path == 'dist' or  dir_path == 'build' or dir_path.endswith('.egg-info'):
                    kpath.remove(os.path.join(root_folder_path, dir_path))

            return


    # Args

    @classmethod
    def has_arg(cls, arg: str) -> bool:
        for passed_arg in cls.get_args(minimum_needed=1):
            if arg.lower() == passed_arg.lower():
                return True

        return False

    @classmethod
    def get_arg(cls, surpress_crash: bool = False) -> Optional[str]:
        args = cls.get_args(minimum_needed=0)

        if not args and not surpress_crash:
            raise ValueError('Argument missing')

        return args[0] if args else None

    @staticmethod
    def get_args(minimum_needed: int = 1) -> List[str]:
        args = sys.argv[1:]

        if len(args) < minimum_needed + 1:
            raise ValueError('Not enough args passed: {} out of {}'.format(len(args), minimum_needed))

        return args


    # Paths

    @classmethod
    def ensure_and_get_path(cls) -> str:
        cwd = os.getcwd()

        if not os.path.exists(cls.setup_file_path()):
            raise ValueError('No \'setup.py\' was found in the current directory')

        return cwd

    @staticmethod
    def create_file(
        file_path: str,
        text: Optional[str] = None,
        overwrite: bool = False
    ) -> bool:
        if not overwrite and os.path.exists(file_path):
            return False

        folder_path = kpath.folder_path_of_file(file_path)

        if not os.path.exists(folder_path):
            os.makedirs(folder_path)

        return strio.save(file_path, text or '')

    @classmethod
    def config_path(cls) -> str:
        return cls.path(Constants.CONFIG_FILE_NAME, cls.home_dir())

    @staticmethod
    def home_dir() -> str:
        return sh('echo $HOME').strip()

    @classmethod
    def readme_path(cls) -> str:
        return os.path.join(os.getcwd(), Constants.README_FILE_NAME)

    @classmethod
    def gitignore_path(cls) -> str:
        return cls.path(Constants.GITIGNORE_FILE_NAME)

    @classmethod
    def install_dependencies_path(cls) -> str:
        return cls.path(Constants.INSTALL_DEPENDENCIES_FILE_NAME)

    @classmethod
    def demo_path(cls) -> str:
        return cls.path(Constants.DEMO_FILE_NAME)

    @classmethod
    def license_path(cls) -> str:
        return cls.path(Constants.LICENSE_FILE_NAME)

    @classmethod
    def setup_file_path(cls) -> str:
        return cls.path(Constants.SETUP_FILE_NAME)

    @classmethod
    def init_file_path(cls, base_path: Optional[str] = None) -> str:
        return cls.path(Constants.INIT_PY_FILE_NAME, base_path=base_path)

    @staticmethod
    def path(file_name: str, base_path: Optional[str] = None) -> str:
        return os.path.join(base_path or os.getcwd(), file_name)


    # Git

    @classmethod
    def get_git_username(cls) -> Optional[str]:
        return cls.__get_local_git_username() or cls.__get_local_git_username_from_git_url() or cls.get_config(True).default_author or cls.__get_global_git_username()

    @classmethod
    def get_git_url(cls) -> Optional[str]:
        url = cls.__get_arg_from_file('.git/config', 'url = ')

        return url.replace('.git', '').strip('/') if url else None

    @classmethod
    def get_git_repo_name(cls) -> Optional[str]:
        git_url = cls.get_git_url()

        return git_url.split('/')[-1] if git_url else None

    @classmethod
    def get_full_git_repo_name(cls) -> Optional[str]:
        git_username = cls.__get_local_git_username_from_git_url()

        if not git_username:
            return None

        git_repo_name = cls.get_git_repo_name()

        if not git_repo_name:
            return None

        return '{}/{}'.format(git_username, git_repo_name)


    # Config

    @classmethod
    def get_config(cls, create_new_if_none: bool = True) -> Optional[Config]:
        config = Config.load(cls.config_path())

        return config if config else cls.__create_new_config() if create_new_if_none else None

    @classmethod
    def __create_new_config(cls) -> Config:
        default_author, default_commit_message, min_v, max_v = Prompt.config(
            default_author=cls.__get_local_git_username() or cls.__get_global_git_username(),
            default_git_message=Constants.DEFAULT_COMMIT_MESSAGE
        )

        config = Config(
            default_author=default_author,
            default_commit_message=default_commit_message,
            default_min_python_version=min_v,
            default_max_python_version=max_v
        )

        config.save(cls.config_path())

        return config


    # Strings

    @classmethod
    def get_paths_name_class(cls, file_name_or_path: str) -> Tuple[str, str, str]:
        file_path = os.path.join(os.getcwd(), file_name_or_path.strip().strip(os.path.sep).strip())
        folder_path = kpath.folder_path_of_file(file_path)

        if not file_path.endswith('.py'):
            file_path += '.py'

        file_name = file_path.split(os.path.sep)[-1].replace('.py', '')
        file_class = cls.class_name_from_file_name(file_name)

        return folder_path, file_path, file_name, file_class

    @staticmethod
    def class_name_from_file_name(file_name: str) -> str:
        return ''.join([p.title() for p in file_name.split('_')])

    @classmethod
    def get_current_package_name(cls) -> str:
        setup = strio.load(cls.setup_file_path())

        return setup.split('name =')[-1].split('name=')[-1].split(',')[0].strip().replace('"', '').replace("'", '')


    # ------------------------------------------------------- Private methods -------------------------------------------------------- #

    # Git

    @classmethod
    def __get_local_git_username(cls) -> Optional[str]:
        return cls.__get_arg_from_file('.git/config', 'name = ')

    @classmethod
    def __get_global_git_username(cls) -> Optional[str]:
        return cls.__get_arg_from_file(os.path.join(cls.home_dir(), '.gitconfig'), 'name = ')

    @classmethod
    def __get_local_git_username_from_git_url(cls) -> Optional[str]:
        git_url = cls.get_git_url()

        return git_url.split('/')[-2] if git_url else None

    @staticmethod
    def __get_arg_from_file(path: str, splitter: str) -> Optional[str]:
        if not os.path.exists(path):
            return None

        try:
            for line in strio.load(path).split('\n'):
                if splitter in line:
                    return line.split(splitter)[1].strip()
        except Exception as e:
            print(e)

        return None


# ---------------------------------------------------------------------------------------------------------------------------------------- #



readme = '''
# [PACKAGE_NAME]

[SHIELDS]

## Description

[DESCRIPTION]

## Install

~~~~bash
pip install [PACKAGE_NAME]
# or
pip3 install [PACKAGE_NAME]
~~~~

## Usage

~~~~python
import [PACKAGE_NAME]
~~~~

## Dependencies

[DEPENDENCIES]
'''.strip()



# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
import os

# Local
from .utils import Utils
from .texts import new_class, new_enum, new_license, file, flow, gitignore, readme, setup

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------ Public methods ------------------------------------------------------------ #

def create_new_class(name: str):
    _, file_path, _, _class = Utils.get_paths_name_class(name)

    Utils.create_file(file_path, new_class(_class))

def create_new_enum(name: str):
    _, file_path, _, _class = Utils.get_paths_name_class(name)

    Utils.create_file(file_path, new_enum(_class))

def create_new_file(name: str):
    _, file_path, _, _ = Utils.get_paths_name_class(name)

    Utils.create_file(file_path, file)

def create_new_flow(name: str):
    _, file_path, _, _ = Utils.get_paths_name_class(name)

    Utils.create_file(file_path, flow)

def create_new_gitignore():
    Utils.create_file(Utils.gitignore_path(), gitignore)

def create_new_readme():
    Utils.create_file(Utils.gitignore_path(), gitignore)

def create_new_subpackage(relative_folder_path: str, create_class: bool = True):
    _, init_file_path, _, _ = Utils.get_paths_name_class(Utils.init_file_path(relative_folder_path))

    if create_class:
        _, _, class_file_name, _class = Utils.get_paths_name_class(relative_folder_path)
        create_new_class(os.path.join(relative_folder_path, class_file_name))

        Utils.create_file(init_file_path, 'from .{} import {}'.format(class_file_name, _class))
    else:
        Utils.create_file(init_file_path, '')


# ---------------------------------------------------------------------------------------------------------------------------------------- #



# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, List
import os, time

# Pip
from kcu import kpath, strio, sh
from kdependencies import Dependencies, InstalledPackage

# Local
from .constants import Constants
from .utils import Utils
from .prompt import Prompt
from .texts import new_api, new_class, new_enum, new_license, file, flow, gitignore, new_readme, updated_readme, new_setup, updated_setup, new_install_dependencies_file, current_version_number

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------- class: Flows ------------------------------------------------------------- #

class Flows:

    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    # Main

    @classmethod
    def new_package(cls, package_name: Optional[str]):
        input('Ensure you\'re in the correct path: \'{}\''.format(os.getcwd()))

        res = Utils.get_args(0)
        package_name = res[0] if res > 0 else None
        config = Utils.get_config(True)
        package_name, description = Prompt.new_package(package_name, os.getcwd().split(os.sep)[-1])

        if not os.path.exists(Utils.setup_file_path()):
            Utils.create_file(Utils.setup_file_path(), new_setup(
                package_name=package_name,
                min_python_version=config.default_min_python_version,
                max_python_version=config.default_max_python_version,
                license_str='License :: OSI Approved :: MIT License'
            ))

        if not os.path.exists(Utils.readme_path()):
            cls.create_new_readme(package_name, Utils.get_full_git_repo_name(), description, open=False)

        if not os.path.exists(Utils.gitignore_path()):
            cls.create_new_gitignore(open=False)

        if not os.path.exists(Utils.demo_path()):
            Utils.create_file(Utils.demo_path())

        cls.create_new_subpackage(os.path.join(package_name, 'models'), create_class=True)
        cls.create_new_subpackage(os.path.join(package_name, 'core'), create_class=True)
        cls.create_new_subpackage(package_name, create_class=True)

    @classmethod
    def upgrade(cls, ensure_path: bool = True, clean_lines: bool = True):
        if ensure_path:
            Utils.ensure_and_get_path()

        if clean_lines:
            cls.clean_lines(ensure_path=False)

        print('Getting dependencies...')
        dependencies = Dependencies.get()

        print('Updating setup.py')
        old_setup_str = strio.load(Utils.setup_file_path())
        updated_setup_str = updated_setup(old_setup_str, dependencies)
        Utils.create_file(Utils.setup_file_path(), updated_setup_str, overwrite=True)

        print('Updating README.md')
        old_readme_str = strio.load(Utils.readme_path())
        demo_str = strio.load(Utils.demo_path())
        updated_readme_str = updated_readme(old_readme_str, demo_str, dependencies)
        Utils.create_file(Utils.readme_path(), updated_readme_str, overwrite=True)

        cls.create_install_file(dependencies, open=False)

    @classmethod
    def publish(cls, ensure_path: bool = True, clean_lines: bool = True, reinstall: bool = True):
        if ensure_path:
            Utils.ensure_and_get_path()

        current_package_name = Utils.get_current_package_name()

        print('Upgrading \'{}\''.format(current_package_name))
        cls.upgrade(ensure_path=False, clean_lines=clean_lines)

        print('Publishing \'{}\' to pypi'.format(current_package_name))
        Utils.publish()

        if reinstall:
            print('Reinstalling \'{}\''.format(current_package_name))
            cls.reinstall(current_package_name, version=current_version_number(Utils.setup_file_path()), max_install_try_count=5)

    @classmethod
    def publish_and_push(cls, message: Optional[str] = None, clean_lines: bool = True):
        cls.publish(ensure_path=True, clean_lines=clean_lines, reinstall=False)
        cls.push(ensure_path=False, clean_lines=False)

        current_package_name = Utils.get_current_package_name()
        print('Reinstalling \'{}\''.format(current_package_name))
        cls.reinstall(current_package_name, version=current_version_number(Utils.setup_file_path()), max_install_try_count=5)

    @classmethod
    def clean_lines(cls, ensure_path: bool = True):
        if ensure_path:
            Utils.ensure_and_get_path()

        for p in kpath.file_paths_from_folder(os.getcwd(), allowed_extensions=cls.__allowed_extesions()):
            with open(p, 'r') as f:
                org_text = f.read()
                text = '\n'.join([l.rstrip() for l in org_text.strip().split('\n')]) + '\n'

            if org_text != text:
                print('Cleaning: \'{}\''.format(p))

                with open(p, 'w') as f:
                    f.write(text)


    # Git

    @classmethod
    def push(cls, message: Optional[str] = None, ensure_path: bool = True, clean_lines: bool = True):
        if ensure_path:
            Utils.ensure_and_get_path()

        if clean_lines:
            cls.clean_lines(ensure_path=False)

        print(sh.sh('git add .', debug=True))
        print(sh.sh('git commit -a -m \'{}\''.format(message or Utils.get_config(True).default_commit_message), debug=True))
        print(sh.sh('git push', debug=True))

    @staticmethod
    def fetch():
        print(sh.sh('git fetch', debug=True))

    @classmethod
    def pull(cls):
        cls.fetch()
        print(sh.sh('git pull', debug=True))


    # Pip install

    @staticmethod
    def uninstall(package: str):
        print(Utils.pip('uninstall -y {}'.format(package)))

    @staticmethod
    def install(package: str, version: Optional[str] = None, max_try_count: Optional[int] = None):
        package_name = '{}{}'.format(package, '=={}'.format(version) if version else '')
        max_try_count = max_try_count or 1
        current_try_count = 0

        while True:
            current_try_count += 1
            res = Utils.pip('install -U {}'.format(package_name))

            if not res.startswith('ERROR'):
                print('Successfully installed \'{}\''.format(package_name))
            elif current_try_count < max_try_count:
                time.sleep(2)
            else:
                break

        print('ERROR: Could not install \'{}\''.format(package_name))

    @classmethod
    def reinstall(cls, package: str, version: Optional[str] = None, max_install_try_count: Optional[int] = None):
        cls.uninstall(package)
        cls.install(package, version, max_try_count=max_install_try_count)


    # New files

    @staticmethod
    def create_install_file(dependencies: Optional[List[InstalledPackage]] = None, open: bool = True):
        file_path = Utils.install_dependencies_path()

        if not dependencies:
            if os.path.exists(file_path):
                os.remove(file_path)

            return

        print('Creating \'{}\''.format(Constants.INSTALL_DEPENDENCIES_FILE_NAME))
        Utils.create_file(file_path, new_install_dependencies_file(dependencies), overwrite=True)

        if open:
            Utils.vscode_open(file_path)

    @staticmethod
    def create_new_api(name: str, open: bool = True):
        _, file_path, _, _class = Utils.get_paths_name_class(name)
        Utils.create_file(file_path, new_api(_class))

        if open:
            Utils.vscode_open(file_path)

    @staticmethod
    def create_new_class(name: str, open: bool = True):
        _, file_path, _, _class = Utils.get_paths_name_class(name)
        Utils.create_file(file_path, new_class(_class))

        if open:
            Utils.vscode_open(file_path)

    @staticmethod
    def create_new_enum(name: str, open: bool = True):
        _, file_path, _, _class = Utils.get_paths_name_class(name)

        Utils.create_file(file_path, new_enum(_class))

        if open:
            Utils.vscode_open(file_path)

    @staticmethod
    def create_new_file(name: str, open: bool = True):
        _, file_path, _, _ = Utils.get_paths_name_class(name)

        Utils.create_file(file_path, file)

        if open:
            Utils.vscode_open(file_path)

    @staticmethod
    def create_new_flow(name: str, open: bool = True):
        _, file_path, _, _ = Utils.get_paths_name_class(name)

        Utils.create_file(file_path, flow)

        if open:
            Utils.vscode_open(file_path)

    @staticmethod
    def create_new_gitignore(open: bool = True):
        Utils.create_file(Utils.gitignore_path(), gitignore)

        if open:
            Utils.vscode_open(Utils.gitignore_path())

    @staticmethod
    def create_new_readme(
        package_name: str,
        full_repo_name: Optional[str] = None,
        description: Optional[str] = None,
        open: bool = False
    ):
        file_path = Utils.readme_path()
        Utils.create_file(file_path, new_readme(package_name=package_name, full_repo_name=full_repo_name, description=description))

        if open:
            Utils.vscode_open(file_path)

    @classmethod
    def create_new_subpackage(cls, relative_folder_path: str, create_class: bool = True):
        _, init_file_path, _, _ = Utils.get_paths_name_class(Utils.init_file_path(relative_folder_path))

        if create_class:
            _, _, class_file_name, _class = Utils.get_paths_name_class(relative_folder_path)
            cls.create_new_class(os.path.join(relative_folder_path, class_file_name), open=False)

            Utils.create_file(init_file_path, 'from .{} import {}'.format(class_file_name, _class))
        else:
            Utils.create_file(init_file_path, '')

    # ------------------------------------------------------ Private properties ------------------------------------------------------ #



    # ------------------------------------------------------- Private methods -------------------------------------------------------- #

    @staticmethod
    def __allowed_extesions() -> List[str]:
        return [
            '.py',
            '.js',
            '.ts',
            '.json'
        ]


# ---------------------------------------------------------------------------------------------------------------------------------------- #



import setuptools, os

readme_path = 'README.md'

if os.path.exists(readme_path):
    with open(readme_path, "r") as f:
        long_description = f.read()
else:
    long_description = 'kcliutils'

setuptools.setup(
    name="kcliutils",
    version="0.0.7",
    author="Kristof",
    description="kcliutils",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kkristof200/py_cli_utils",
    packages=setuptools.find_packages(),
    install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            # Main

            'new_python_package=kcliutils.__main__:new_package',
            'npp=kcliutils.__main__:new_package',

            'upgrade_python_package=kcliutils.__main__:upgrade',
            'upp=kcliutils.__main__:upgrade',

            'publish_python_package=kcliutils.__main__:publish',
            'ppp=kcliutils.__main__:publish',

            'publish_and_push_python_package=kcliutils.__main__:publish_and_push',
            'pipush=kcliutils.__main__:publish_and_push',
            'pppp=kcliutils.__main__:publish_and_push',

            'clean_lines=kcliutils.__main__:clean_lines',
            'cl=kcliutils.__main__:clean_lines',


            # Git

            'push=kcliutils.__main__:push',
            'psh=kcliutils.__main__:push',

            'fetch=kcliutils.__main__:fetch',
            'ftch=kcliutils.__main__:fetch',

            'pull=kcliutils.__main__:pull',
            'pll=kcliutils.__main__:pull',


            # Pip

            'pip_uninstall=kcliutils.__main__:uninstall',
            'pipu=kcliutils.__main__:uninstall',

            'pip_install=kcliutils.__main__:install',
            'pipi=kcliutils.__main__:install',
            'pipiu=kcliutils.__main__:install',

            'pip_reinstall=kcliutils.__main__:reinstall',
            'pipir=kcliutils.__main__:reinstall',


            # New files

            'new_python_install_file=kcliutils.__main__:create_install_file',
            'npif=kcliutils.__main__:create_install_file',

            'new_python_api=kcliutils.__main__:create_new_api',
            'npa=kcliutils.__main__:create_new_api',

            'new_python_class=kcliutils.__main__:create_new_class',
            'npc=kcliutils.__main__:create_new_class',

            'new_python_enum=kcliutils.__main__:create_new_enum',
            'npe=kcliutils.__main__:create_new_enum',

            'new_python_file=kcliutils.__main__:create_new_file',
            'npf=kcliutils.__main__:create_new_file',

            'new_python_flow=kcliutils.__main__:create_new_flow',
            'npfl=kcliutils.__main__:create_new_flow',

            'new_python_subpackage=kcliutils.__main__:create_new_subpackage',
            'nps=kcliutils.__main__:create_new_subpackage'
        ]
    },
    python_requires='>=3.4',
)



from .api import api
from .class_ import class_
from .enum_ import enum_
from .file import file
from .flow import flow
from .gitignore import gitignore
from .license import license
from .readme import readme
from .setup import setup
from .install_dependencies import install_dependencies



# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, List

# Pip
from kdependencies import InstalledPackage
from kcu import strings

# Local
from .core_texts import readme
from .utils import multi_replace

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------ Public methods ------------------------------------------------------------ #

def updated_readme(
    readme: str,
    demo: Optional[str],
    dependencies: Optional[List[InstalledPackage]]
) -> str:
    if demo:
        usage = strings.between(readme, '## Usage', '##')

        if usage:
            readme.replace(usage.strip(), '\n\n~~~~python\n{}\n~~~~\n\n'.format(demo.strip()))

    if dependencies:
        old_dependencies = strings.between(readme, '## Dependencies', '##')

        if old_dependencies:
            readme.replace(old_dependencies.strip(), ', '.join(['[{}]({})'.format(d.name, 'https://pypi.org/project/{}'.format(d.name) if not d.private else d.home_url) for d in dependencies]))

    return readme.strip()

def new_readme(
    package_name: str,
    full_repo_name: Optional[str] = None,
    description: Optional[str] = None,
    dependencies: Optional[List[InstalledPackage]] = None
) -> str:
    return multi_replace(
        readme,
        {
            '[PACKAGE_NAME]': package_name,
            '[SHIELDS]': __create_shields(package_name, full_repo_name) if full_repo_name else '',
            '[DESCRIPTION]': description or '',
            '[DEPENDENCIES]': ', '.join(['[{}]({})'.format(d.name, 'https://pypi.org/project/{}'.format(d.name) if not d.private else d.home_url) for d in dependencies]) if dependencies else ''
        }
    )


# ----------------------------------------------------------- Private methods ------------------------------------------------------------ #

def __create_shields(
    package_name: str,
    full_repo_name: Optional[int]
):
    return ' '.join(
        [
            __shield('GitHub - last commit', 'github/last-commit/{}'.format(full_repo_name)),
            __shield('GitHub - commit activity', 'github/commit-activity/m/{}'.format(full_repo_name)),
            __shield('GitHub - code size in bytes', 'github/languages/code-size/{}'.format(full_repo_name)),
            __shield('GitHub - repo size', 'github/repo-size/{}'.format(full_repo_name)),
            __shield('GitHub - lines of code', 'tokei/lines/github/{}'.format(full_repo_name)),
            __shield('GitHub - license', 'github/license/{}'.format(full_repo_name)),

            __shield('PyPI - Python Version', 'pypi/pyversions/{}?logo=pypi'.format(package_name)),
            __shield('PyPI - Downloads', 'pypi/dm/{}?logo=pypi'.format(package_name))
        ]
    )

def __shield(name: str, url_data: str, style: str = 'flat-square') -> str:
    return '![{}](https://img.shields.io/{}{}style={})'.format(name, url_data, '&' if '?' in url_data else '?', style)


# ---------------------------------------------------------------------------------------------------------------------------------------- #
